from smartcard.scard import SCARD_CTL_CODE, SCARD_SHARE_DIRECT
from smartcard.System import readers
from smartcard.util import toHexString
from smartcard.Exceptions import NoCardException
from smartcard.CardConnection import CardConnection
from smartcard.CardMonitoring import CardMonitor, CardObserver
from colorama import init, Fore, Style


def list_readers():
    pcsc_readers = readers()
    if len(pcsc_readers) == 0:
        print("No smart card readers found.")
    else:
        print("Available readers:")
        for index, reader in enumerate(pcsc_readers):
            print(f"{index + 1}: {reader}")
    return pcsc_readers

def connect_to_reader(reader):
    try:
        connection = reader.createConnection()
        connection.connect()
        return connection, False
    except NoCardException:
        # Try to connect in direct mode
        connection.connect(protocol=CardConnection.RAW_protocol, mode=SCARD_SHARE_DIRECT)
        print(Style.BRIGHT + Fore.LIGHTMAGENTA_EX + "Connected using Direct Mode" + Style.RESET_ALL)
        return connection, True
    except Exception as e:
        print(Fore.RED + f"Exception: {e}" + Style.RESET_ALL)

    return None

def transmit_data(connection, data):
    apdu = [int(x, 16) for x in data.split()]
    response, sw1, sw2 = connection.transmit(apdu)
    print(Style.BRIGHT + Fore.YELLOW + f"Response: {toHexString(response)}")
    print(f"Status words: {sw1:02X} {sw2:02X}" + Style.RESET_ALL)

def control_data(connection, data):
    command_data = [int(x, 16) for x in data.split()]
    response_data = connection.control(SCARD_CTL_CODE(3500), command_data)
    print(Style.BRIGHT + Fore.YELLOW + f"Response: {toHexString(response_data)}")
    print(f"Status words: {response_data[len(response_data) - 2]:02X} {response_data[len(response_data) - 1]:02X}" + Style.RESET_ALL)

class CardObserverImpl(CardObserver):
    def update(self, observable, actions):
        (addedcards, removedcards) = actions
        for card in addedcards:
            print(Fore.LIGHTGREEN_EX + f"\n{card.reader}: Card present" + Style.RESET_ALL)
            print(toHexString(card.atr))
        for card in removedcards:
            print(Fore.LIGHTRED_EX + f"\n{card.reader}: Card absent"  + Style.RESET_ALL)

if __name__ == "__main__":
    card_monitor = CardMonitor()
    observer = CardObserverImpl()
    card_monitor.addObserver(observer)
    
    while True:
        pcsc_readers = list_readers()
        user_input = input(Style.BRIGHT + Fore.CYAN + "Select the reader to connect or press 'q' to exit: " + Style.RESET_ALL)
        if user_input.lower() == 'q':
            break
        reader_index = int(user_input, 10) - 1
        connection, direct = connect_to_reader(pcsc_readers[reader_index])
        if connection:
            if direct:
                data = input("Enter the control data to transmit (in hex, e.g., 'FF 00 01 00 00'): ")
                control_data(connection, data)
            else:
                data = input("Enter the APDU to transmit (in hex, e.g., '80 84 00 00 08'): ")
                transmit_data(connection, data)
            connection.disconnect()
    card_monitor.deleteObserver(observer)